#include <cstdlib>
#include <cctype>
#include <cstdio>
#include <cerrno>
#include <cstring>
#include <string>
#include <list>

#include "parser.h"

struct sh_error       { std::string cmd; int signaled; };
struct auto_error     {};
struct incomplete_run {};

static void system1(const char *cmd);
static void system2(const char *cmd);
static void check_run_format(int runs, const char *Run);
static void find_cmds_make(char * auto_dir);

// visible within this file for auto_error& catch block
static char fname[10000], qrun[10000], auto_dir[10000];;

int main(int argc, const char *argv[]) try {
	// mandatory arguments
	const char *Name, *Run;
	// optional arguments
	const char *RestartRun=0;
	// default start index of optional rfile arguments
	int Options=3, estat;
	bool Force = false, Modified, Incomplete, RFExists, LFExists;
	
	// show usage if no arguments are given
	if(argc<Options) return print_usage();
	
	// show usage if switch '-h' was given
	if( strcmp("-h", argv[1])==0 ) return print_usage();
	
	// look for AUTO's cmds.make file
	find_cmds_make(auto_dir);
	
	// check optional switch '-f'
	if( (Force = (strcmp("-f", argv[1])==0)) ) ++Options;
	
	// read mandatory arguments <Name> and <Run>, or
	// show usage if too few arguments are given
	if(argc<Options) return print_usage();
	Name = argv[Options-2];
	Run  = argv[Options-1];
	
	// read optional argument <RestartRun>
	if(argc>Options && argv[Options][0]!='-') {
		RestartRun=argv[Options++];
		sprintf(fname, "data/q.%s", RestartRun);
		FILE *h = fopen(fname, "r");
		if(h==NULL || fclose(h)==EOF)
			throw cstdlib_error(errno, fname);
	}
	
	// read default constants file r.<Name> or data/r.<RestartRun>
	if(RestartRun != NULL)
		sprintf(fname, "data/r.%s", RestartRun);
	else
		sprintf(fname, "r.%s", Name);
	read_constants_file(fname);
	
	// parse command line for rfile options
	read_rauto_args(Options, argc, argv);
	
	// create data directory if necessary
	system2("test -d data || mkdir data");

	// loop over all restart labels
	std::list<std::string> irsl = IRS_list();
	std::list<std::string>::iterator it=irsl.begin(), end=irsl.end();
	int runs = irsl.size(), curr_run = 1;
	
	// check if command line option IRS
	// a) was present for restarted runs
	// b) was absent or `-IRS 0' for initial runs
	if(RestartRun!=NULL) {
		if(runs<=1 && *it == "0") throw arg_error(1, "IRS",
			"restart label mising, option -IRS must be given for restarted runs"
		);
	} else {
		if(runs>1 || *it != "0") throw parse_error(
			"option -IRS must not be present for restarted runs"
		);
	}
	
	// check if Run contains valid format specifier
	check_run_format(runs, Run);

	// be a bit verbose
	printf("rauto: *** number of runs=%d\n", runs);
	
	for( ; it!=end; ++it) {
		// Run may contain a %d format specifier, compute qualified run name
		int lab = atoi(it->c_str());
		sprintf(qrun, Run, lab);
		
		// be a bit more verbose
		printf("rauto: *** starting run %s: run %d of %d, IRS=%s\n",
			qrun, curr_run++, runs, it->c_str());

		// write auto constants to file fort.2
		write_constants_file(*it);
		
		if(RestartRun!=NULL) {
			// write printls file to file l.<qrun>
			sprintf(fname, "printls %s %d > l.%s", RestartRun, lab, qrun);
			system2(fname);
		}

		// perform up-to-date tests if '-f' switch was not given
		if( !Force ) {
			// check if an old constants file already exists
			sprintf(fname, "test -f data/r.%s", qrun);
			estat = system(fname);
			if( WIFSIGNALED(estat) ) return 1;
			RFExists = (WEXITSTATUS(estat)==0);

			if( RFExists ) {
				// check if equations file was modified after last run
				sprintf(fname, "test %s.f -nt data/r.%s", Name, qrun);
				estat = system(fname);
				if( WIFSIGNALED(estat) ) return 1;
				Modified = (WEXITSTATUS(estat)==0);

				// check if the new constants file is different from the old one
				sprintf(fname, "cmp -s fort.2 data/r.%s", qrun);
				estat = system(fname);
				if( WIFSIGNALED(estat) ) return 1;
				Modified = Modified || (WEXITSTATUS(estat)!=0);
			} else {
				Modified = true;
			}
			
			if(RestartRun != NULL) {
				// check if an old printls file exists
				sprintf(fname, "test -f data/l.%s", qrun);
				estat = system(fname);
				if( WIFSIGNALED(estat) ) return 1;
				LFExists = (WEXITSTATUS(estat)==0);

				if( LFExists ) {
					// check if the new printls file is different from the old one
					sprintf(fname, "cmp -s l.%s data/l.%s", qrun, qrun);
					estat = system(fname);
					if( WIFSIGNALED(estat) ) return 1;
					Modified = Modified || (WEXITSTATUS(estat)!=0);
				}
			}
		}

		if(Force || Modified) {
			if(RestartRun!=NULL) {
				// copy restart data to fort.3 for every iteration,
				// because AUTO deletes fort.3
				sprintf(fname, "cp -f data/q.%s fort.3", RestartRun);
				system2(fname);
			} else {
				// make sure there is an empty fort.3 for initial runs
				sprintf(fname, "rm -f fort.3 ; touch fort.3");
				system2(fname);
			}

			// save copy of constants file, because AUTO deletes fort.2
			sprintf(fname, "cp -f fort.2 r.%s", qrun);
			system2(fname);

			// run auto
			sprintf(fname, "%s%s%s ; %s%s%s%s",
				"export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:", auto_dir, "/lib",
				"make -f ", auto_dir, "/cmds.make EQUATION_NAME=", Name);
			system1(fname);

			// check for complete execution of the current run
			sprintf(fname, "test -f fort.7 && test -s fort.7");
			estat = system(fname);
			if( WIFSIGNALED(estat) ) return 1;
			Incomplete = (WEXITSTATUS(estat)!=0);
			sprintf(fname, "test -f fort.8 && test -s fort.8");
			estat = system(fname);
			if( WIFSIGNALED(estat) ) return 1;
			Incomplete = Incomplete || (WEXITSTATUS(estat)!=0);
			sprintf(fname, "test -f fort.9 && test -s fort.9");
			estat = system(fname);
			if( WIFSIGNALED(estat) ) return 1;
			Incomplete = Incomplete || (WEXITSTATUS(estat)!=0);
			
			if( Incomplete ) throw incomplete_run();

			// save [pqdrl].* files
			sprintf(fname, "mv -f fort.7 data/p.%s", qrun);
			system2(fname);
			sprintf(fname, "mv -f fort.8 data/q.%s", qrun);
			system2(fname);
			sprintf(fname, "mv -f fort.9 data/d.%s", qrun);
			system2(fname);
			sprintf(fname, "mv -f r.%s data/r.%s", qrun, qrun);
			system2(fname);
			sprintf(fname, "mv -f l.%s data/l.%s", qrun, qrun);
			if(RestartRun != NULL) system2(fname);
		} else {
			printf("rauto: *** run %s up to date, nothing done\n", qrun);
		}
		
		// clean up
		sprintf(fname, "rm -f fort.2 fort.3 fort.7 fort.8 fort.9 [rl].%s", qrun);
		system2(fname);
	}

	return 0;

} catch(parse_error &err) {
	err.print("rauto");
	return 1;

} catch(auto_error &) {
	sprintf(fname, "rm -f fort.2 fort.3 fort.7 fort.8 fort.9 [rl].%s", qrun);
	fprintf(stderr, "%s\n", fname);
	system(fname);
	return 1;

} catch(incomplete_run &) {
	fprintf(stderr, "%s%s%s\n",
		"rauto: *** computation of run ",
		qrun,
		" was incomplete, no data saved");
	sprintf(fname, "rm -f fort.2 fort.3 fort.7 fort.8 fort.9 [rl].%s", qrun);
	system(fname);
	return 1;

} catch(sh_error &err) {
	if(err.signaled) return 1;
	fprintf(stderr,
		"rauto: shell returned with error, offending command was:\n");
	fprintf(stderr, "%s\n", err.cmd.c_str());
	return 1;
}


/*******************************************************************
***                       local functions                       ****
*******************************************************************/

static void system1(const char *cmd) {
	#ifndef NDEBUG
		printf("%s\n", cmd);
	#else
		int estat = system(cmd);
		if(WEXITSTATUS(estat) || WIFSIGNALED(estat))
			throw auto_error();
	#endif
}

static void system2(const char *cmd) {
	#ifndef NDEBUG
		printf("%s\n", cmd);
	#else
		int estat = system(cmd);
    if(WEXITSTATUS(estat) || WIFSIGNALED(estat)) {
			sh_error err;
			err.cmd = cmd;
			err.signaled = WIFSIGNALED(estat);
			throw err;
		}
	#endif
}

static void check_run_format(int runs, const char *Run) {
	unsigned int count = 0, slen = strlen(Run);

	// check if invalid characters are present
	const char * inv_chars="./|<>?*&[]{} `'";
	if(slen != strcspn(Run, inv_chars)) {
		std::string ermsg = "run name must not contain any of the characters:\n";
		ermsg += inv_chars;
		throw parse_error(ermsg);
	}
	
	// format: *%[0[W]]d*
	for(unsigned int i=0; i<slen; ++i) if(Run[i]=='%') ++count;
	if( runs==1 && count==0 ) return;
	if( count>1 )
		throw parse_error("RUN must not contain more than one %d format specifier");
	if( runs>1 && count!=1 )
		throw parse_error("in multiple run mode RUN must contain exactly one %d format specifier");
	
	// check if %d specifier has correct syntax
	unsigned int i=0;
	const char *msg = "invalid format specifier, accepted format is %[0[W]]d";
	while(Run[i++]!='%');
	if(Run[i] == 'd') return;
	if(Run[i] != '0') throw parse_error(msg);
	while(isdigit(Run[++i]));
	if(Run[i] == 'd') return;
	throw parse_error(msg);
}

static void find_cmds_make(char * auto_dir) {
	FILE *h;
	const char *RAUTO_DIR = getenv("RAUTO_DIR");
	const char *AUTO_DIR  = getenv("AUTO_DIR") ;
	const char *CAUTO_DIR = AUTODIR ;
	const char *dirf = "%s/cmds.make", *subdirf = "%s/cmds/cmds.make",
		*dirpathf = "%s", *subdirpathf = "%s/cmds";
	
	// check if RAUTO_DIR is set
	if( RAUTO_DIR!=NULL ) {
		sprintf(auto_dir, dirf, RAUTO_DIR);
		h=fopen(auto_dir, "r");
		sprintf(auto_dir, dirpathf, RAUTO_DIR);
		if(h!=NULL) { fclose(h); return; }

		sprintf(auto_dir, subdirf, RAUTO_DIR);
		h=fopen(auto_dir, "r");
		sprintf(auto_dir, subdirpathf, RAUTO_DIR);
		if(h!=NULL) { fclose(h); return; }
	}
	
	// check if AUTO_DIR is set
	if( AUTO_DIR!=NULL ) {
		sprintf(auto_dir, dirf, AUTO_DIR);
		h=fopen(auto_dir, "r");
		sprintf(auto_dir, dirpathf, AUTO_DIR);
		if(h!=NULL) { fclose(h); return; }

		sprintf(auto_dir, subdirf, AUTO_DIR);
		h=fopen(auto_dir, "r");
		sprintf(auto_dir, subdirpathf, AUTO_DIR);
		if(h!=NULL) { fclose(h); return; }
	}

	// check compile-time path
	sprintf(auto_dir, dirf, CAUTO_DIR);
	h=fopen(auto_dir, "r");
	sprintf(auto_dir, dirpathf, CAUTO_DIR);
	if(h!=NULL) { fclose(h); return; }
	
	sprintf(auto_dir, subdirf, CAUTO_DIR);
	h=fopen(auto_dir, "r");
	sprintf(auto_dir, subdirpathf, CAUTO_DIR);
	if(h!=NULL) { fclose(h); return; }
	
	throw parse_error("could not find file cmds.make, set AUTO_DIR or RAUTO_DIR to the correct location");
}
