#include <cctype>
#include <cstdio>
#include <cerrno>
#include <cstring>
#include <string>
#include <list>
#include <map>
#include <fstream>

#include "scanner.h"
#include "parser.h"

/*******************************************************************
***            implementation part of scanner.h                 ****
*******************************************************************/

std::string argstream::catArgs(int arg, int argc, const char *argv[]) {
	std::string s;
	for(int a=arg; a<argc; ++a) {
		s += argv[a];
		s += ' ';
	}
	return s;
}

static tokens read_num(argstream &);
static tokens read_fracpart(tokens t, argstream &);
static tokens read_name(argstream &);

tokens scanner::peek() {
	if(!Peek) { Token=get(); Peek=true; }
	return Token;
}

tokens scanner::get() {
	if(Peek) {Peek=false; return Token;}
	
	tokens t;
	int c=In.peek();

	while(c == ' ' || c == '\t') {c=In.get(); c=In.peek();}
	
	if(isdigit(c)) return read_num(In);
	
	switch(c) {
		case -1:
			t.type=In.get();
			return t;

		case ':':
			t.type=In.get();
			t.val = ':';
			return t;
		
		case '-':
			c=In.get();
			c=In.peek();
			if(isdigit(c)) {
				t = read_num(In);
				t.val = '-' + t.val;
			} else if (isalpha(c)) {
				t = read_name(In);
			} else if (isspace(c)) {
				t.type = '-';
			} else if (c=='.') {
				t.val = "-.";
				c=In.get(); c=In.peek();
				if(!isdigit(c)) break;
				t = read_fracpart(t, In);
			} else {
				break;
			}
			return t;
		
		 case '+':
			c=In.get();
			c=In.peek();
			if(isdigit(c)) {
				t = read_num(In);
				t.val = t.val;
			} else if (isspace(c)) {
				t.type = '+';
			} else {
				break;
			}
			return t;

		 case '.':
			c=In.get();
			if(isdigit(In.peek())) {
				t = read_fracpart(t, In);
				t.val = '.' + t.val;
			} else {
				break;
			}
			return t;
	}
	
	t.type = tokens::ERR;
	return t;
}

tokens read_num(argstream & In) {
	tokens t;
	t.type = tokens::INUM;
	t.val  = In.get();
	
	while(isdigit(In.peek())) t.val += In.get();
	if(In.peek() == '.') {
		t.val += In.get();
		return read_fracpart(t, In);
	} else if (In.peek() == 'e') {
		return read_fracpart(t, In);
	} else if(!isspace(In.peek()) && In.peek()!=':')
		t.type = tokens::ERR;
	
	return t;
}

tokens read_fracpart(tokens t, argstream & In) {
	t.type = tokens::DNUM;
	
	while(isdigit(In.peek())) t.val += In.get();
	if(toupper(In.peek()) == 'E') {
		t.val += In.get();
		if(In.peek() == '+' || In.peek() == '-') t.val += In.get();
		if(!isdigit(In.peek())) {
			t.type = tokens::ERR;
			return t;
		}
		while(isdigit(In.peek())) t.val += In.get();
	}
	if(!isspace(In.peek()) && In.peek()!=':') t.type = tokens::ERR;
	return t;
}

tokens read_name(argstream & In) {
	tokens t;
	t.type = tokens::ONAME;
	t.val  = toupper(In.get());
	
	while(isalnum(In.peek())) t.val += toupper(In.get());
	if(!isspace(In.peek())) t.type = tokens::ERR;
	return t;
}

/*******************************************************************
***             implementation part of parser.h                 ****
*******************************************************************/

struct rfent {
	struct vlent { std::string par, val, inc, eval; };
	enum { IVAL, DVAL, ILST1, ILST2, IDLST };
	
	int type;
	bool acceptRange, acceptEmpty, acceptDirec;
	std::list<vlent> vals;
};

typedef std::list<rfent::vlent>::iterator rfentit;

static  std::map<std::string,rfent> RFile;
typedef std::map<std::string,rfent>::iterator RFileIterator;

static struct {
	const char *Name;
	int         Type;
	bool        AcceptRange, AcceptEmpty, InitEmpty, AcceptDirec;
	const char *IVal, *Comment;

} const RFileEntries[] = {

	{ "NDIM",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "IPS",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "IRS",   rfent::ILST1, true,  false, false, false, "0" , NULL },
	{ "ILP",   rfent::IVAL,  false, false, false, false, NULL, "NDIM,IPS,IRS,ILP" },

	{ "ICP",   rfent::ILST2, false, false, false, false, NULL, "NICP,(ICP(I),I=1,NICP)" },

	{ "NTST",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "NCOL",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "IAD",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "ISP",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "ISW",   rfent::IVAL,  false, false, false, false, "1" , NULL },
	{ "IPLT",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "NBC",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "NINT",  rfent::IVAL,  false, false, false, false, NULL, "NTST,NCOL,IAD,ISP,ISW,IPLT,NBC,NINT" },
	
	{ "NMX",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "RL0",   rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "RL1",   rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "A0",    rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "A1",    rfent::DVAL,  false, false, false, false, NULL, "NMX,RL0,RL1,A0,A1" },
	
	{ "NPR",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "MXBF",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "IID",   rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "ITMX",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "ITNW",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "NWTN",  rfent::IVAL,  false, false, false, false, NULL, NULL },
	{ "JAC",   rfent::IVAL,  false, false, false, false, NULL, "NPR,MXBF,IID,ITMX,ITNW,NWTN,JAC" },
	
	{ "EPSL",  rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "EPSU",  rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "EPSS",  rfent::DVAL,  false, false, false, false, NULL, "EPSL,EPSU,EPSS" },
	
	{ "DS",    rfent::DVAL,  false, false, false, true , NULL, NULL },
	{ "DSMIN", rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "DSMAX", rfent::DVAL,  false, false, false, false, NULL, NULL },
	{ "IADS",  rfent::IVAL,  false, false, false, false, NULL, "DS,DSMIN,DSMAX,IADS" },
	
	{ "THL",   rfent::IDLST, false, true , false, false, NULL, "NTHL,((I,THL(I)),I=1,NTHL)" },
	{ "THU",   rfent::IDLST, false, true , false, false, NULL, "NTHU,((I,THU(I)),I=1,NTHU)" },
	{ "UZR",   rfent::IDLST, true , true , true , false, NULL, "NUZR,((I,UZR(I)),I=1,NUZR)" },
	
	{ NULL }
};

static void print_full_usage();
static void print_arg_usage(std::string argname);

int print_usage(std::string argname) {
	if(argname == "")
		print_full_usage();
	else
		print_arg_usage(argname);
	return 1;
}

static const char *argf    = "%s%s",   *repargf    = "%s ...%s",
	 /* *optargf = "[%s]%s", */ *optrepargf = "[%s ...]%s",
	 *optgrprepargf = "[(%s) ...]%s",
	 *inum = "INUM", *dnum = "DNUM", *dnumdir="(DNUM|+|-)",
	 *irange = "IRANGE", /* *drange = "DRANGE", */
	 *idnum = "INUM DNUM", *idrange = "INUM DRANGE",
	 *noexp     = "",
	 *irangeexp = ", IRANGE = INUM | INUM:[INUM:]INUM",
	 *drangeexp = ", DRANGE = DNUM | DNUM:[DNUM:]DNUM",
	 *indent = "       |";

void print_full_usage() {
	const char *format, *argtype, *explanation;
	bool noindent = true, noprint = true;
	int printcount = 0, cols = 8;

	fprintf(stderr, "%s%s\n",
		"usage: rauto [-h|-f] NAME RUN ",
		"[RESTART_RUN -IRS RLABS] [OPTIONS ...]");

	fprintf(stderr, "RLABS  : ");
	fprintf(stderr, repargf, irange, irangeexp);
	fprintf(stderr, "\n");

	fprintf(stderr, "OPTIONS: %s | %s | %s | %s\n",
		"(-IOPT INUM)", "(-DOPT DNUM)", "SOPT", "LOPT");

	fprintf(stderr, "IOPT   :");
	noindent   = true;
	printcount = 0;
	format     = argf;
	
	for(int i=0; RFileEntries[i].Name!=NULL; ++i) {
		if(RFileEntries[i].Type == rfent::IVAL) {
			if(!noindent) fprintf(stderr, "%s", indent);
			if(printcount > 0) fprintf(stderr, " |");
			noindent = true;
			fprintf(stderr, " %-4s", RFileEntries[i].Name);
			if(++printcount > cols) {
				fprintf(stderr, "\n");
				noindent = false;
				printcount = 0;
			}
		}
	}
	if(0<printcount && printcount<=cols) fprintf(stderr, "\n");

	fprintf(stderr, "DOPT   :");
	noindent   = true;
	printcount = 0;
	format     = argf;
	
	for(int i=0; RFileEntries[i].Name!=NULL; ++i) {
		if(RFileEntries[i].Type == rfent::DVAL && !RFileEntries[i].AcceptDirec) {
			if(!noindent) fprintf(stderr, "%s", indent);
			if(printcount > 0) fprintf(stderr, " |");
			noindent = true;
			fprintf(stderr, " %-4s", RFileEntries[i].Name);
			if(++printcount > cols) {
				fprintf(stderr, "\n");
				noindent = false;
				printcount = 0;
			}
		}
	}
	if(0<printcount && printcount<=cols) fprintf(stderr, "\n");
	
	fprintf(stderr, "SOPT   :");
	noindent = true;
	noprint  = true;
	
	for(int i=0; RFileEntries[i].Name!=NULL; ++i) {
		switch(RFileEntries[i].Type) {
			case rfent::DVAL:
				if(RFileEntries[i].AcceptDirec) {
					format      = argf;
					argtype     = dnum;
					explanation = noexp;
					argtype     = dnumdir;
					noprint     = false;
				}
				break;

			case rfent::ILST1:
			case rfent::ILST2:
				if( strcmp(RFileEntries[i].Name, "IRS")==0 ) break;
				format      = repargf;
				argtype     = inum;
				explanation = noexp;
				if(RFileEntries[i].AcceptEmpty) format = optrepargf;
				if(RFileEntries[i].AcceptRange) {
					argtype     = irange;
					explanation = irangeexp;
				}
				noprint = false;
				break;

		}
		if(!noprint) {
			if(!noindent) fprintf(stderr, "%s", indent);
			fprintf(stderr, " -%-4s", RFileEntries[i].Name);
			fprintf(stderr, format, argtype, explanation);
			fprintf(stderr, "\n");
			noprint  = true;
			noindent = false;
		}
	}

	fprintf(stderr, "LOPT   :");
	noindent = true;
	noprint  = true;

	for(int i=0; RFileEntries[i].Name!=NULL; ++i) {
		if(RFileEntries[i].Type == rfent::IDLST) {
			format      = repargf;
			argtype     = idnum;
			explanation = noexp;
			if(RFileEntries[i].AcceptEmpty) format = optgrprepargf;
			if(RFileEntries[i].AcceptRange) {
				argtype     = idrange;
				explanation = drangeexp;
			}
			if(!noindent) fprintf(stderr, "%s", indent);
			fprintf(stderr, " -%-4s", RFileEntries[i].Name);
			fprintf(stderr, format, argtype, explanation);
			fprintf(stderr, "\n");
			noindent = false;
		}
	}

}

void print_arg_usage(std::string argname) {
	const char *format, *argtype, *explanation;

	fprintf(stderr, "syntax: ");
	
	for(int i=0; RFileEntries[i].Name!=NULL; ++i) {
		if(argname==RFileEntries[i].Name) {
			switch(RFileEntries[i].Type) {
				case rfent::IVAL:
					format      = argf;
					argtype     = inum;
					explanation = noexp;
					break;

				case rfent::DVAL:
					format      = argf;
					argtype     = dnum;
					explanation = noexp;
					if(RFileEntries[i].AcceptDirec) argtype = dnumdir;
					break;

				case rfent::ILST1:
				case rfent::ILST2:
					format      = repargf;
					argtype     = inum;
					explanation = noexp;
					if(RFileEntries[i].AcceptEmpty) format = optrepargf;
					if(RFileEntries[i].AcceptRange) {
						argtype     = irange;
						explanation = irangeexp;
					}
					break;

				case rfent::IDLST:
					format      = repargf;
					argtype     = idnum;
					explanation = noexp;
					if(RFileEntries[i].AcceptEmpty) format = optrepargf;
					if(RFileEntries[i].AcceptRange) {
						argtype     = idrange;
						explanation = drangeexp;
					}
					break;

			}
			fprintf(stderr, "-%s ", RFileEntries[i].Name);
			fprintf(stderr, format, argtype, explanation);
			fprintf(stderr, "\n");
		}
	}
}

void skip(std::ifstream & istr) {
	int c=istr.get();
	while(c!='\n' && c!=EOF) c=istr.get();
}

void read_constants_file(const char * fname) {
	static const char *rerr = "could not read entry";
	int row=1, col=1, i=0;
	std::ifstream in(fname);
	if(!in) throw stl_error(0,0, fname, "could not open file for reading");
	
	while(RFileEntries[i].Name != NULL) {
		switch(RFileEntries[i].Type) {
			case rfent::IVAL:
			case rfent::DVAL:
			case rfent::ILST1: {
				rfent::vlent vent;
				in >> vent.val;
				if(!in) throw stl_error(row, col, fname, rerr, RFileEntries[i].Name);
				++col;
				if(RFileEntries[i].IVal != NULL) {
					vent.val = RFileEntries[i].IVal;
				}
				if(RFileEntries[i].AcceptDirec) {
					unsigned int i;
					for(i=0; i<vent.val.size(); ++i) {
						if(vent.val[i]=='+' || vent.val[i]=='-') {
							vent.val = vent.val.substr(i+1);
							break;
						}
					}
				}
				rfent ent;
				ent.vals.push_back(vent);
				ent.type = RFileEntries[i].Type;
				ent.acceptRange = RFileEntries[i].AcceptRange;
				ent.acceptEmpty = RFileEntries[i].AcceptEmpty;
				ent.acceptDirec = RFileEntries[i].AcceptDirec;
				RFile[RFileEntries[i].Name] = ent;
				if(RFileEntries[i].Comment != NULL) {
					skip(in); 
					++row;
					col=1;
				}
				break;
			}
			
			case rfent::ILST2: {
				int length;
				in >> length;
				if(!in) throw stl_error(row, col, fname, rerr, RFileEntries[i].Name);
				++col;
				rfent ent;
				for(int j=0; j<length; ++j) {
					rfent::vlent vent;
					in >> vent.val;
					if(!in) throw stl_error(row, col, fname, rerr, RFileEntries[i].Name);
					++col;
					if( ! RFileEntries[i].InitEmpty )
						ent.vals.push_back(vent);
				}
				ent.type = RFileEntries[i].Type;
				ent.acceptRange = RFileEntries[i].AcceptRange;
				ent.acceptEmpty = RFileEntries[i].AcceptEmpty;
				ent.acceptDirec = RFileEntries[i].AcceptDirec;
				RFile[RFileEntries[i].Name] = ent;
				if(RFileEntries[i].Comment != NULL) {
					skip(in); 
					++row;
					col=1;
				}
				break;
			}
				
			case rfent::IDLST: {
				int length;
				in >> length;
				if(!in) throw stl_error(row, col, fname, rerr, RFileEntries[i].Name);
				++col;
				if(RFileEntries[i].Comment != NULL) {
					skip(in); 
					++row;
					col=1;
				}
				rfent ent;
				for(int j=0; j<length; ++j) {
					rfent::vlent vent;
					in >> vent.par >> vent.val;
					if(!in) throw stl_error(row, col, fname, rerr, RFileEntries[i].Name);
					++row;
					if( ! RFileEntries[i].InitEmpty )
						ent.vals.push_back(vent);
				}
				ent.type = RFileEntries[i].Type;
				ent.acceptRange = RFileEntries[i].AcceptRange;
				ent.acceptEmpty = RFileEntries[i].AcceptEmpty;
				ent.acceptDirec = RFileEntries[i].AcceptDirec;
				RFile[RFileEntries[i].Name] = ent;
				break;
			}
		}
	
	++i;
	}
}

static rfent expand(rfent & ent);

void write_constants_file(std::string IRS) {
	const char *rerr = "could not write entry";
	const char *fname = "fort.2";
	int i = 0;
	std::ofstream out(fname);
	if(!out) throw stl_error(0,0, fname, "could not open file for writing");
	
	while(RFileEntries[i].Name != NULL) {
		switch(RFileEntries[i].Type) {
			case rfent::IVAL:
			case rfent::DVAL: {
				rfent ent = expand(RFile[RFileEntries[i].Name]);
				rfentit it = ent.vals.begin();
				out << ' ' << it->val;
				if(!out) throw stl_error(0,0, fname, rerr, RFileEntries[i].Name);
				if(RFileEntries[i].Comment != NULL)
					out << ' ' << RFileEntries[i].Comment << std::endl;
				break;
			}
			
			case rfent::ILST1: {
				out << ' ' << IRS;
				if(!out) throw stl_error(0,0, fname, rerr, RFileEntries[i].Name);
				break;
			}
			
			case rfent::ILST2: {
				rfent ent = expand(RFile[RFileEntries[i].Name]);
				out << ' ' << ent.vals.size();
				if(!out) throw stl_error(0,0, fname, rerr, RFileEntries[i].Name);
				rfentit it, end = ent.vals.end();
				for(it=ent.vals.begin(); it!=end; ++it) {
					out << ' ' << it->val;
					if(!out) throw stl_error(0,0, fname, rerr, RFileEntries[i].Name);
				}
				if(RFileEntries[i].Comment != NULL)
					out << ' ' << RFileEntries[i].Comment << std::endl;
				break;
			}
				
			case rfent::IDLST: {
				rfent ent = expand(RFile[RFileEntries[i].Name]);
				out << ' ' << ent.vals.size();
				if(RFileEntries[i].Comment != NULL)
					out << ' ' << RFileEntries[i].Comment << std::endl;
				if(!out) throw stl_error(0,0, fname, rerr, RFileEntries[i].Name);
				rfentit it, end = ent.vals.end();
				for(it=ent.vals.begin(); it!=end; ++it) {
					out << ' ' << it->par << ' ' << it->val << std::endl;
					if(!out) throw stl_error(0,0, fname, rerr, RFileEntries[i].Name);
				}
				break;
			}
		}
	
	++i;
	}
}

std::list<std::string> IRS_list() {
	rfent ent = expand(RFile["IRS"]);
	std::list<std::string> irsl;
	rfentit it, end = ent.vals.end();
	for(it=ent.vals.begin(); it!=end; ++it) irsl.push_back(it->val);
	return irsl;
}

static rfent expand_iranges(rfent & ent);
static rfent expand_dranges(rfent & ent);

rfent expand(rfent & ent) {
	if(!ent.acceptRange) return ent;
	switch(ent.type) {
		case rfent::ILST1:
		case rfent::ILST2:
			return expand_iranges(ent);
		
		case rfent::IDLST:
			return expand_dranges(ent);
		
		case rfent::IVAL: case rfent::DVAL: default:
			throw parse_error("cannot expand IVALs or DVALs, this is a programming error");
	}
}

rfent expand_iranges(rfent & ent) {
	char nbuff[100];
	rfent xent;
	xent.type = ent.type;
	xent.acceptRange = ent.acceptRange;
	rfentit it, end = ent.vals.end();
	
	for(it=ent.vals.begin(); it!=end; ++it) {
		if(it->inc=="") {
			xent.vals.push_back(*it);
		} else {
			int bval=atoi(it->val.c_str()),
					vinc=atoi(it->inc.c_str()),
					eval=atoi(it->eval.c_str());
			std::string par = it->par;
			if(vinc>0) {
				for(int val=bval; val<=eval; val+=vinc) {
					sprintf(nbuff, "%d", val);
					rfent::vlent vent;
					vent.par = par;
					vent.val = nbuff;
					xent.vals.push_back(vent);
				}
			} else if(vinc<0) {
				for(int val=bval; val>=eval; val+=vinc) {
					sprintf(nbuff, "%d", val);
					rfent::vlent vent;
					vent.par = par;
					vent.val = nbuff;
					xent.vals.push_back(vent);
				}
			}
		}
	}

	return xent;
}

rfent expand_dranges(rfent & ent) {
	char nbuff[1000];
	rfent xent;
	xent.type = ent.type;
	xent.acceptRange = ent.acceptRange;
	rfentit it, end = ent.vals.end();
	
	for(it=ent.vals.begin(); it!=end; ++it) {
		if(it->inc=="") {
			xent.vals.push_back(*it);
		} else {
			double bval=atof(it->val.c_str()),
						 vinc=atof(it->inc.c_str()),
						 eval=atof(it->eval.c_str())+0.1*vinc;
			std::string par = it->par;
			if(vinc > 0.0) {
				for(double val=bval; val<=eval; val+=vinc) {
					sprintf(nbuff, "%.16e", val);
					rfent::vlent vent;
					vent.par = par;
					vent.val = nbuff;
					xent.vals.push_back(vent);
				}
			} else if(vinc < 0.0) {
				for(double val=bval; val>=eval; val+=vinc) {
					sprintf(nbuff, "%.16e", val);
					rfent::vlent vent;
					vent.par = par;
					vent.val = nbuff;
					xent.vals.push_back(vent);
				}
			}
		}
	}
	
	return xent;
}

/*******************************************************************
***   entry point for parsing command line arguments of rauto   ****
*******************************************************************/

static int read_entry_data(std::string OName, rfent &ent, scanner &s);

void read_rauto_args(int arg, int argc, const char *argv[]) {
	argstream   in(arg, argc, argv);
	scanner     s(in);
	std::string OName = "";
	int         Arg;
	
	for(tokens t=s.peek(); t.type!=-1; t=s.peek()) {
		switch(t.type) {
			case tokens::ONAME: {
				RFileIterator RIt = RFile.find(t.val);
				if(RIt == RFile.end()) throw invalid_arg(t.val);
				t     = s.get();
				OName = t.val;
				rfent &ent = RIt->second;
				Arg=read_entry_data(OName, ent, s);
				break;
			}
			
			case -1: return;
			
			default:
				if(OName=="")
					throw invalid_arg(t.val);
				else
					throw too_many_args_error(Arg+1, OName);
		}
	}
}

static void read_irange(int arg, std::string OName,
	rfent::vlent &vent, scanner &s);
static void read_drange(int arg, std::string OName,
	rfent::vlent &vent, scanner &s);

int read_entry_data(std::string OName, rfent &ent, scanner &s) {
	int Arg = 1;
	tokens t = s.peek();
	switch(ent.type) {
		case rfent::IVAL: {
			if(t.type != tokens::INUM) throw arg_error(Arg, OName);
			rfent::vlent vent;
			if(ent.acceptRange) {
				read_irange(Arg, OName, vent, s);
			} else {
				t = s.get();
				vent.val = t.val;
			}
			ent.vals.clear();
			ent.vals.push_back(vent);
			break;
		}
		
		case rfent::DVAL: {
			rfent::vlent vent;
			if(ent.acceptDirec) {
				switch(t.type) {
					case '+':
					case '-':
						t = s.get();
						rfentit it = ent.vals.begin();
						std::string num = it->val;
						unsigned int i;
						for(i=0; i<num.size(); ++i) {
							if(num[i]=='+' || num[i]=='-') {
								num = num.substr(i+1);
								break;
							}
						}
						num = (t.type=='-') ? "-"+num : num;
						vent.val = num;
						goto exit_dval;
				}
			}
			if(t.type != tokens::INUM && t.type != tokens::DNUM)
				throw arg_error(Arg, OName);
			if(ent.acceptRange) {
				read_drange(Arg, OName, vent, s);
			} else {
				t = s.get();
				vent.val = t.val;
			}
			exit_dval:
			ent.vals.clear();
			ent.vals.push_back(vent);
			break;
		}
		
		case rfent::ILST1:
		case rfent::ILST2: {
			if(t.type != tokens::INUM)
				if(ent.acceptEmpty) {
					ent.vals.clear();
					break;
				} else
					throw arg_error(Arg, OName);
			rfent::vlent vent;
			if(ent.acceptRange) {
				read_irange(Arg, OName, vent, s);
			} else {
				t = s.get();
				vent.val = t.val;
			}
			ent.vals.clear();
			ent.vals.push_back(vent);
			for(t=s.peek(); t.type==tokens::INUM; t=s.peek()) {
				vent.inc = "";
				if(ent.acceptRange) {
					read_irange(++Arg, OName, vent, s);
				} else {
					t = s.get();
					++Arg;
					vent.val = t.val;
				}
				ent.vals.push_back(vent);
			}
			break;
		}
		
		case rfent::IDLST: {
			if(t.type != tokens::INUM)
				if(ent.acceptEmpty) {
					ent.vals.clear();
					break;
				} else
					throw arg_error(Arg, OName);
			t = s.get();
			rfent::vlent vent;
			vent.par = t.val;
			t = s.peek();
			if(t.type != tokens::INUM && t.type != tokens::DNUM)
				throw arg_error(Arg+1, OName);
			if(ent.acceptRange) {
				read_drange(++Arg, OName, vent, s);
			} else {
				t = s.get();
				++Arg;
				vent.val = t.val;
			}
			ent.vals.clear();
			ent.vals.push_back(vent);
			for(t=s.peek(); t.type==tokens::INUM; t=s.peek()) {
				t = s.get();
				++Arg;
				vent.par = t.val;
				t = s.peek();
				if(t.type != tokens::INUM && t.type != tokens::DNUM)
					throw arg_error(Arg+1, OName);
				vent.inc = "";
				if(ent.acceptRange) {
					read_drange(++Arg, OName, vent, s);
				} else {
					t = s.get();
					++Arg;
					vent.val = t.val;
				}
				ent.vals.push_back(vent);
			}
			break;
		}
		
	}
	return Arg;
}

void read_irange(int Arg, std::string OName, rfent::vlent &vent, scanner &s) {
	tokens t = s.get();
	vent.val = t.val;
	t = s.peek();
	if(t.type != ':') return;
	vent.inc  = "1";
	t = s.get();
	t = s.peek();
	if(t.type != tokens::INUM) throw arg_error(Arg, OName);
	t = s.get();
	vent.eval = t.val;
	t = s.peek();
	if(t.type != ':') return;
	vent.inc = vent.eval;
	t = s.get();
	t = s.peek();
	if(t.type != tokens::INUM) throw arg_error(Arg, OName);
	t = s.get();
	vent.eval = t.val;
}

void read_drange(int Arg, std::string OName, rfent::vlent &vent, scanner &s) {
	tokens t = s.get();
	vent.val = t.val;
	t = s.peek();
	if(t.type != ':') return;
	vent.inc  = "1.0";
	t = s.get();
	t = s.peek();
	if(t.type != tokens::INUM && t.type != tokens::DNUM)
		throw arg_error(Arg, OName);
	t = s.get();
	vent.eval = t.val;
	t = s.peek();
	if(t.type != ':') return;
	vent.inc = vent.eval;
	t = s.get();
	t = s.peek();
	if(t.type != tokens::INUM && t.type != tokens::DNUM)
		throw arg_error(Arg, OName);
	t = s.get();
	vent.eval = t.val;
}

void parse_error::print(const char *pname) {
	if(!msg.empty()) fprintf(stderr, "%s: %s\n", pname, msg.c_str());
}

void arg_error::print(const char *pname) {
	parse_error::print(pname);
	fprintf(stderr, "%s: syntax error in argument %d of %s, or argument missing\n",
		pname, Arg, OName.c_str());
	print_usage(OName);
}

void too_many_args_error::print(const char *pname) {
	fprintf(stderr, "%s: %s %d of %s, or too many arguments\n",
		pname, "syntax error near argument", Arg, OName.c_str());
	print_usage(OName);
}

void invalid_arg::print(const char *pname) {
	fprintf(stderr, "unrecognised argument \"%s\"\n", token.c_str());
}

void cstdlib_error::print(const char *pname) {
	fprintf(stderr, "%s:%s:%s\n", pname, msg.c_str(), strerror(errnum));
}

void stl_error::print(const char *pname) {
	if(row==0)
		if(entry == "")
			fprintf(stderr, "%s:%s:%s\n",
				pname, fname.c_str(), msg.c_str());
		else
			fprintf(stderr, "%s:%s:%s %s\n",
				pname, fname.c_str(), msg.c_str(), entry.c_str());
	else
		fprintf(stderr, "%s:%s:%d,%d:%s %s\n",
			pname, fname.c_str(), row, col, msg.c_str(), entry.c_str());
}
